"""Base classes for IBM watsonx.ai large language models."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import (  # type: ignore[import-untyped]
    Model,
    ModelInference,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from ibm_watsonx_ai.metanames import (  # type: ignore[import-untyped]
    GenTextParamsMetaNames,
)
from langchain_core.language_models.llms import BaseLLM
from langchain_core.outputs import Generation, GenerationChunk, LLMResult
from langchain_core.utils.utils import secret_from_env
from pydantic import AliasChoices, ConfigDict, Field, SecretStr, model_validator
from typing_extensions import Self

from langchain_ibm.utils import (
    async_gateway_error_handler,
    extract_params,
    gateway_error_handler,
    normalize_api_key,
    resolve_watsonx_credentials,
    secret_from_env_multi,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator, Mapping

    from langchain_core.callbacks import (
        AsyncCallbackManagerForLLMRun,
        CallbackManagerForLLMRun,
    )

logger = logging.getLogger(__name__)
textgen_valid_params = [
    value for key, value in GenTextParamsMetaNames.__dict__.items() if key.isupper()
]


class WatsonxLLM(BaseLLM):
    """`IBM watsonx.ai` large language models class.

    ???+ info "Setup"

        To use the large language models, you need to have the `langchain_ibm` python
        package installed, and the environment variable `WATSONX_API_KEY` set with your
        API key or pass it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        ```python
        from langchain_ibm import WatsonxLLM
        from ibm_watsonx_ai.metanames import GenTextParamsMetaNames

        parameters = {
            GenTextParamsMetaNames.DECODING_METHOD: "sample",
            GenTextParamsMetaNames.MAX_NEW_TOKENS: 100,
            GenTextParamsMetaNames.MIN_NEW_TOKENS: 1,
            GenTextParamsMetaNames.TEMPERATURE: 0.5,
            GenTextParamsMetaNames.TOP_K: 50,
            GenTextParamsMetaNames.TOP_P: 1,
        }

        model = WatsonxLLM(
            model_id="google/flan-t5-xl",
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",
            params=parameters,
            # api_key="*****"
        )
        ```

    ??? info "Invoke"

        ```python
        input_text = "The meaning of life is "
        response = model.invoke(input_text)
        print(response)
        ```

        ```txt
        "42, but what was the question?
        The answer to the ultimate question of life, the universe, and everything is 42.
        But what was the question? This is a reference to Douglas Adams' science fiction
        series "The Hitchhiker's Guide to the Galaxy."
        ```

    ??? info "Stream"

        ```python
        for chunk in model.stream(input_text):
            print(chunk, end="")
        ```

        ```txt
        "42, but what was the question?
        The answer to the ultimate question of life, the universe, and everything is 42.
        But what was the question? This is a reference to Douglas Adams' science fiction
        series "The Hitchhiker's Guide to the Galaxy."
        ```

    ??? info "Async"
        ```python
        response = await model.ainvoke(input_text)

        # stream:
        # async for chunk in model.astream(input_text):
        #     print(chunk, end="")

        # batch:
        # await model.abatch([input_text])
        ```

        ```txt
        "42, but what was the question?
        The answer to the ultimate question of life, the universe, and everything is 42.
        But what was the question? This is a reference to Douglas Adams' science fiction
        series "The Hitchhiker's Guide to the Galaxy."
        ```
    """

    model_id: str | None = None
    """Type of model to use."""

    model: str | None = None
    """
    Name or alias of the foundation model to use.
    When using IBM's watsonx.ai Model Gateway (public preview), you can specify any
    supported third-party model-OpenAI, Anthropic, NVIDIA, Cerebras, or IBM's own
    Granite series—via a single, OpenAI-compatible interface. Models must be explicitly
    provisioned (opt-in) through the Gateway to ensure secure, vendor-agnostic access
    and easy switch-over without reconfiguration.

    For more details on configuration and usage, see [IBM watsonx Model Gateway docs](https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-gateway.html?context=wx&audience=wdp)
    """

    deployment_id: str | None = None
    """Type of deployed model to use."""

    project_id: str | None = None
    """ID of the Watson Studio project."""

    space_id: str | None = None
    """ID of the Watson Studio space."""

    url: SecretStr = Field(
        alias="url",
        default_factory=secret_from_env("WATSONX_URL", default=None),  # type: ignore[assignment]
    )
    """URL to the Watson Machine Learning or CPD instance."""

    apikey: SecretStr | None = None
    api_key: SecretStr | None = Field(
        default_factory=secret_from_env_multi(
            names_priority=["WATSONX_API_KEY", "WATSONX_APIKEY"],
            deprecated={"WATSONX_APIKEY"},
        ),
        serialization_alias="api_key",
        validation_alias=AliasChoices("api_key", "apikey"),  # accept both on input
        description="API key to the Watson Machine Learning or CPD instance.",
    )
    """API key to the Watson Machine Learning or CPD instance."""

    token: SecretStr | None = Field(
        alias="token",
        default_factory=secret_from_env("WATSONX_TOKEN", default=None),
    )
    """Token to the CPD instance."""

    password: SecretStr | None = Field(
        alias="password",
        default_factory=secret_from_env("WATSONX_PASSWORD", default=None),
    )
    """Password to the CPD instance."""

    username: SecretStr | None = Field(
        alias="username",
        default_factory=secret_from_env("WATSONX_USERNAME", default=None),
    )
    """Username to the CPD instance."""

    instance_id: SecretStr | None = Field(
        alias="instance_id",
        default_factory=secret_from_env("WATSONX_INSTANCE_ID", default=None),
    )
    """Instance_id of the CPD instance."""

    version: SecretStr | None = None
    """Version of the CPD instance."""

    params: dict[str, Any] | None = None
    """Model parameters to use during request generation."""

    verify: str | bool | None = None
    """You can pass one of following as verify:
        * the path to a CA_BUNDLE file
        * the path of directory with certificates of trusted CAs
        * True - default path to truststore will be taken
        * False - no verification will be made"""

    streaming: bool = False
    """ Whether to stream the results or not. """

    watsonx_model: ModelInference = Field(default=None, exclude=True)  #: :meta private:

    watsonx_model_gateway: Gateway = Field(
        default=None,
        exclude=True,
    )  #: :meta private:

    watsonx_client: APIClient | None = Field(default=None)

    model_config = ConfigDict(
        extra="forbid",
    )

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Is lc serializable."""
        return False

    @property
    def lc_secrets(self) -> dict[str, str]:
        """Mapping of secret environment variables."""
        return {
            "url": "WATSONX_URL",
            "api_key": "WATSONX_API_KEY",  # preferred
            "apikey": "WATSONX_APIKEY",
            "token": "WATSONX_TOKEN",
            "password": "WATSONX_PASSWORD",
            "username": "WATSONX_USERNAME",
            "instance_id": "WATSONX_INSTANCE_ID",
        }

    @model_validator(mode="before")
    @classmethod
    def _normalize_and_warn_deprecated_input(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Handle deprecated input kwarg name `apikey` vs new `api_key`.
            data = normalize_api_key(data=data)
        return data

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that credentials and python package exists in environment."""
        if self.watsonx_model_gateway is not None:
            error_msg = (
                "Passing the 'watsonx_model_gateway' parameter to the WatsonxLLM "
                "constructor is not supported yet.",
            )
            raise NotImplementedError(error_msg)

        if isinstance(self.watsonx_model, (ModelInference, Model)):
            self.model_id = self.watsonx_model.model_id
            self.deployment_id = getattr(self.watsonx_model, "deployment_id", "")
            self.project_id = self.watsonx_model._client.default_project_id  # noqa: SLF001
            self.space_id = self.watsonx_model._client.default_space_id  # noqa: SLF001
            self.params = self.watsonx_model.params

        elif isinstance(self.watsonx_client, APIClient):
            if sum(map(bool, (self.model, self.model_id, self.deployment_id))) != 1:
                error_msg = (
                    "The parameters 'model', 'model_id' and 'deployment_id' are "
                    "mutually exclusive. Please specify exactly one of these "
                    "parameters when initializing WatsonxLLM.",
                )
                raise ValueError(error_msg)
            if self.model is not None:
                watsonx_model_gateway = Gateway(
                    api_client=self.watsonx_client,
                    verify=self.verify,
                )
                self.watsonx_model_gateway = watsonx_model_gateway
            else:
                watsonx_model = ModelInference(
                    model_id=self.model_id,
                    deployment_id=self.deployment_id,
                    params=self.params,
                    api_client=self.watsonx_client,
                    project_id=self.project_id,
                    space_id=self.space_id,
                    verify=self.verify,
                )
                self.watsonx_model = watsonx_model

        else:
            if sum(map(bool, (self.model, self.model_id, self.deployment_id))) != 1:
                error_msg = (
                    "The parameters 'model', 'model_id' and 'deployment_id' are "
                    "mutually exclusive. Please specify exactly one of these "
                    "parameters when initializing WatsonxLLM.",
                )
                raise ValueError(error_msg)

            credentials = resolve_watsonx_credentials(
                url=self.url,
                api_key=self.api_key,
                token=self.token,
                password=self.password,
                username=self.username,
                instance_id=self.instance_id,
                version=self.version,
                verify=self.verify,
            )
            if self.model is not None:
                watsonx_model_gateway = Gateway(
                    credentials=credentials,
                    verify=self.verify,
                )
                self.watsonx_model_gateway = watsonx_model_gateway
            else:
                watsonx_model = ModelInference(
                    model_id=self.model_id,
                    deployment_id=self.deployment_id,
                    credentials=credentials,
                    params=self.params,
                    project_id=self.project_id,
                    space_id=self.space_id,
                )
                self.watsonx_model = watsonx_model

        return self

    @gateway_error_handler
    def _call_model_gateway(
        self, *, model: str, prompt: str | list[str] | list[int], **params: Any
    ) -> Any:
        return self.watsonx_model_gateway.completions.create(
            model=model,
            prompt=prompt,
            **params,
        )

    @async_gateway_error_handler
    async def _acall_model_gateway(
        self,
        *,
        model: str,
        prompt: str | list[str] | list[int],
        **params: Any,
    ) -> Any:
        return await self.watsonx_model_gateway.completions.acreate(
            model=model,
            prompt=prompt,
            **params,
        )

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {
            "model_id": self.model_id,
            "deployment_id": self.deployment_id,
            "params": self.params,
            "project_id": self.project_id,
            "space_id": self.space_id,
        }

    @property
    def _llm_type(self) -> str:
        """Return the type of LLM."""
        return "IBM watsonx.ai"

    @staticmethod
    def _extract_token_usage(
        response: list[dict[str, Any]] | None = None,
    ) -> dict[str, Any]:
        if response is None:
            return {"completion_tokens": 0, "prompt_tokens": 0, "total_tokens": 0}

        completion_tokens = 0
        prompt_tokens = 0

        def get_count_value(key: str, result: dict[str, Any]) -> int:
            return result.get(key, 0) or 0

        for res in response:
            results = res.get("results")
            if results:
                prompt_tokens += get_count_value("input_token_count", results[0])
                completion_tokens += get_count_value(
                    "generated_token_count",
                    results[0],
                )
        total_tokens = completion_tokens + prompt_tokens
        return {
            "completion_tokens": completion_tokens,
            "prompt_tokens": prompt_tokens,
            "total_tokens": total_tokens,
        }

    @staticmethod
    def _validate_chat_params(
        params: dict[str, Any],
    ) -> dict[str, Any]:
        """Validate and fix the chat parameters."""
        for param in params:
            if param.lower() not in textgen_valid_params:
                error_msg = (
                    f"Parameter {param} is not valid. "
                    f"Valid parameters are: {textgen_valid_params}"
                )
                raise ValueError(error_msg)
        return params

    @staticmethod
    def _override_chat_params(
        params: dict[str, Any],
        **kwargs: Any,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Override class parameters with those provided in the invoke method.

        Merges the 'params' dictionary with any 'params' found in kwargs,
        then updates 'params' with matching keys from kwargs and removes
        those keys from kwargs.
        """
        for key in list(kwargs.keys()):
            if key.lower() in textgen_valid_params:
                params[key] = kwargs.pop(key)
        return params, kwargs

    def _get_chat_params(
        self,
        stop: list[str] | None = None,
        **kwargs: Any,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        params = extract_params(kwargs, self.params)

        params, kwargs = self._override_chat_params(params or {}, **kwargs)
        if stop is not None:
            if params and "stop_sequences" in params:
                error_msg = (
                    "`stop_sequences` found in both the input and default params."
                )
                raise ValueError(error_msg)
            params = (params or {}) | {"stop_sequences": stop}
        return params, kwargs

    def _create_llm_result(self, response: list[dict[str, Any]]) -> LLMResult:
        """Create the LLMResult from the choices and prompts."""
        generations = [
            [
                Generation(
                    text=result.get("generated_text", ""),
                    generation_info={"finish_reason": result.get("stop_reason")}
                    | (
                        {"moderations": moderations}
                        if (moderations := result.get("moderations"))
                        else {}
                    ),
                ),
            ]
            for res in response
            if (results := res.get("results"))
            for result in results
        ]
        llm_output = {
            "token_usage": self._extract_token_usage(response),
            "model_id": self.model_id,
            "deployment_id": self.deployment_id,
        }
        return LLMResult(generations=generations, llm_output=llm_output)

    def _create_llm_gateway_result(self, response: dict[str, Any]) -> LLMResult:
        """Create the LLMResult from the choices and prompts."""
        choices = response["choices"]

        generations = [
            [
                Generation(
                    text=choice["text"],
                    generation_info={
                        "finish_reason": choice.get("finish_reason"),
                        "logprobs": choice.get("logprobs"),
                    },
                ),
            ]
            for choice in choices
        ]

        llm_output = {
            "token_usage": response["usage"]["total_tokens"],
            "model_id": self.model_id,
            "deployment_id": self.deployment_id,
        }
        return LLMResult(generations=generations, llm_output=llm_output)

    def _stream_response_to_generation_chunk(
        self,
        stream_response: dict[str, Any],
    ) -> GenerationChunk:
        """Convert a stream response to a generation chunk."""
        result = stream_response.get("results", [{}])[0]
        if not result:
            return GenerationChunk(text="")

        finish_reason = result.get("stop_reason")
        finish_reason = None if finish_reason == "not_finished" else finish_reason

        generation_info = {
            "finish_reason": finish_reason,
            "llm_output": {
                "model_id": self.model_id,
                "deployment_id": self.deployment_id,
            },
        }

        if moderations := result.get("moderations"):
            generation_info["moderations"] = moderations

        return GenerationChunk(
            text=result.get("generated_text", ""),
            generation_info=generation_info,
        )

    def _call(
        self,
        prompt: str,
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> str:
        """Call the IBM watsonx.ai inference endpoint.

        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating the response.
            run_manager: Optional callback manager.
            kwargs: Additional keyword args

        Returns:
            The string generated by the model.

        Example:
            ```python
            response = model.invoke("What is a molecule")
            ```
        """
        result = self._generate(
            prompts=[prompt],
            stop=stop,
            run_manager=run_manager,
            **kwargs,
        )
        return result.generations[0][0].text

    async def _acall(
        self,
        prompt: str,
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> str:
        """Async version of the _call method."""
        result = await self._agenerate(
            prompts=[prompt],
            stop=stop,
            run_manager=run_manager,
            **kwargs,
        )
        return result.generations[0][0].text

    def _generate(
        self,
        prompts: list[str],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        *,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> LLMResult:
        """Call the IBM watsonx.ai inference endpoint that then generates the response.

        Args:
            prompts: List of strings (prompts) to pass into the model
            stop: Optional list of stop words to use when generating the response
            run_manager: Optional callback manager
            stream: Stream response
            kwargs: Additional keyword args

        Returns:
            The full LLMResult output.

        Example:
            ```python
            response = model.generate(["What is a molecule"])
            ```
        """
        params, kwargs = self._get_chat_params(stop=stop, **kwargs)
        params = self._validate_chat_params(params)
        should_stream = stream if stream is not None else self.streaming
        if should_stream:
            if len(prompts) > 1:
                error_msg = (
                    f"WatsonxLLM currently only supports single prompt, got {prompts}"
                )
                raise ValueError(error_msg)
            generation = GenerationChunk(text="")
            stream_iter = self._stream(
                prompts[0],
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            )
            for chunk in stream_iter:
                if generation is None:
                    generation = chunk  # type: ignore[unreachable]
                else:
                    generation += chunk
            if generation is None:
                error_msg = "No generation chunks were received from the stream."  # type: ignore[unreachable]
                raise RuntimeError(error_msg)
            if isinstance(generation.generation_info, dict):
                llm_output = generation.generation_info.pop("llm_output")
                return LLMResult(generations=[[generation]], llm_output=llm_output)
            return LLMResult(generations=[[generation]])
        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **params}
            response = self._call_model_gateway(
                model=self.model,
                prompt=prompts,
                **call_kwargs,
            )
            return self._create_llm_gateway_result(response)
        response = self.watsonx_model.generate(prompt=prompts, params=params, **kwargs)
        return self._create_llm_result(response)

    async def _agenerate(
        self,
        prompts: list[str],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        *,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> LLMResult:
        """Async run the LLM on the given prompt and input."""
        params, kwargs = self._get_chat_params(stop=stop, **kwargs)
        params = self._validate_chat_params(params)
        if stream:
            return await super()._agenerate(
                prompts=prompts,
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            )
        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **params}
            responses = await self._acall_model_gateway(
                model=self.model,
                prompt=prompts,
                **call_kwargs,
            )
            return self._create_llm_gateway_result(responses)
        responses = [
            await self.watsonx_model.agenerate(prompt=prompt, params=params, **kwargs)
            for prompt in prompts
        ]

        return self._create_llm_result(responses)

    def _stream(
        self,
        prompt: str,
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Call the IBM watsonx.ai inference endpoint that then streams the response.

        Args:
            prompt: The prompt to pass into the model
            stop: Optional list of stop words to use when generating the response
            run_manager: Optional callback manager
            kwargs: Additional keyword args

        Returns:
            The iterator which yields generation chunks

        Example:
            ```python
            response = model.stream("What is a molecule")
            for chunk in response:
                print(chunk, end="", flush=True)
            ```
        """
        params, kwargs = self._get_chat_params(stop=stop, **kwargs)
        params = self._validate_chat_params(params)
        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **params, "stream": True}
            chunk_iter = self._call_model_gateway(
                model=self.model,
                prompt=prompt,
                **call_kwargs,
            )
        else:
            chunk_iter = self.watsonx_model.generate_text_stream(
                prompt=prompt,
                params=params,
                **(kwargs | {"raw_response": True}),
            )
        for stream_resp in chunk_iter:
            if not isinstance(stream_resp, dict):
                stream_data = stream_resp.dict()
            else:
                stream_data = stream_resp
            chunk = self._stream_response_to_generation_chunk(stream_data)

            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)
            yield chunk

    async def _astream(
        self,
        prompt: str,
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[GenerationChunk]:
        params, kwargs = self._get_chat_params(stop=stop, **kwargs)
        params = self._validate_chat_params(params)

        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **params, "stream": True}
            chunk_iter = await self._acall_model_gateway(
                model=self.model,
                prompt=prompt,
                **call_kwargs,
            )
        else:
            chunk_iter = await self.watsonx_model.agenerate_stream(
                prompt=prompt,
                params=params,
            )
        async for stream_resp in chunk_iter:
            if not isinstance(stream_resp, dict):
                stream_data = stream_resp.dict()
            else:
                stream_data = stream_resp
            chunk = self._stream_response_to_generation_chunk(stream_data)

            if run_manager:
                await run_manager.on_llm_new_token(chunk.text, chunk=chunk)
            yield chunk

    def get_num_tokens(self, text: str) -> int:
        """Get num tokens."""
        if self.watsonx_model_gateway is not None:
            error_msg = (
                "Tokenize endpoint is not supported by IBM Model Gateway endpoint."
            )
            raise NotImplementedError(error_msg)
        response = self.watsonx_model.tokenize(text, return_tokens=False)
        return cast("int", response["result"]["token_count"])

    def get_token_ids(self, text: str) -> list[int]:
        """Get token ids."""
        error_msg = "API does not support returning token ids."
        raise NotImplementedError(error_msg)



import json
import os
import re
from typing import Any, Literal, cast

import pytest
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    TextChatParameters,
)
from ibm_watsonx_ai.metanames import (  # type: ignore[import-untyped]
    GenTextParamsMetaNames,
)
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    HumanMessage,
    SystemMessage,
    ToolCallChunk,
)
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.tools import tool
from pydantic import BaseModel

from langchain_ibm import ChatWatsonx
from langchain_ibm.agent_toolkits.utility import WatsonxToolkit

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

MODEL_ID = "ibm/granite-3-3-8b-instruct"
MODEL_ID_TOOL = "meta-llama/llama-3-3-70b-instruct"
MODEL_ID_TOOL_2 = "mistralai/mistral-small-3-1-24b-instruct-2503"
MODEL_ID_REASONING_CONTENT = "openai/gpt-oss-120b"

PARAMS_WITH_MAX_TOKENS = {"max_tokens": 20}

CREATE_CHAT_WATSONX_INIT_PARAMETERS = [
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only api_key",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only apikey",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="api_key and apikey",
    ),
]


@pytest.mark.parametrize("init_data", CREATE_CHAT_WATSONX_INIT_PARAMETERS)
def test_chat_watsonx_init(init_data: dict[str, Any]) -> None:
    chat = ChatWatsonx(**init_data)

    messages = [
        ("user", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages)
    assert response
    assert response.content


@pytest.mark.token_check
def test_chat_invoke() -> None:
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    messages = [
        ("user", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages)
    assert response
    assert response.content


def test_chat_invoke_with_reasoning_content() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID_REASONING_CONTENT,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={
            "include_reasoning": True,
            "reasoning_effort": "low",
        },
    )
    messages = [("human", "Say hello!")]
    response = chat.invoke(messages)
    assert response
    assert response.content
    assert response.additional_kwargs.get("reasoning_content")

    response_2 = chat.invoke(messages, params={"include_reasoning": False})
    assert response_2
    assert response_2.content
    assert not response_2.additional_kwargs.get("reasoning_content")


def test_chat_invoke_with_params_as_dict_in_invoke() -> None:
    params = {"max_tokens": 10}
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    messages = [
        ("system", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages, params=params)
    assert response
    assert response.content
    print(response.content)


def test_chat_invoke_with_params_as_object_in_invoke() -> None:
    params = TextChatParameters(max_tokens=10)
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    messages = [
        ("system", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages, params=params)
    assert response
    assert response.content
    print(response.content)


def test_chat_invoke_with_params_as_object_in_constructor() -> None:
    params = TextChatParameters(max_tokens=10)
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params,
    )
    messages = [
        ("system", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages)
    assert response
    assert response.content
    print(response.content)


def test_chat_invoke_with_invoke_params() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
    }
    parameters_2 = {
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    messages = [
        ("system", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = chat.invoke(messages, params=parameters_2)
    assert response
    assert response.content


def test_chat_generate_with_few_inputs() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=PARAMS_WITH_MAX_TOKENS,
    )
    message = HumanMessage(content="Hello")
    response = chat.generate([[message], [message]])
    assert response
    for generation in response.generations:
        assert generation[0].text


def test_chat_generate_with_reasoning_content() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID_REASONING_CONTENT,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={
            "include_reasoning": True,
            "reasoning_effort": "low",
        },
    )
    message = HumanMessage(content="Hello")
    response = chat.generate([[message], [message]])
    assert response
    for generation in response.generations:
        assert generation[0].text
        assert generation[0].message  # type: ignore[attr-defined]
        assert "reasoning_content" in generation[0].message.additional_kwargs  # type: ignore[attr-defined]
        assert generation[0].message.additional_kwargs["reasoning_content"]  # type: ignore[attr-defined]

    response_2 = chat.generate(
        [[message], [message]], params={"include_reasoning": False}
    )
    assert response_2
    for generation in response_2.generations:
        assert generation[0].text
        assert generation[0].message  # type: ignore[attr-defined]
        assert "reasoning_content" not in generation[0].message.additional_kwargs  # type: ignore[attr-defined]


async def test_chat_agenerate() -> None:
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    message = HumanMessage(content="Hello")
    response = await chat.agenerate([[message], [message]])
    assert response
    for generation in response.generations:
        assert generation[0].text


def test_chat_invoke_with_few_various_inputs() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=PARAMS_WITH_MAX_TOKENS,
    )
    system_message = SystemMessage(content="You are to chat with the user.")
    human_message = HumanMessage(content="Hello")
    response = chat.invoke([system_message, human_message])
    assert response
    assert isinstance(response, BaseMessage)
    assert response.content
    assert isinstance(response.content, str)


async def test_chat_ainvoke() -> None:
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    messages = [
        ("user", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    response = await chat.ainvoke(messages)
    assert response
    assert response.content


def test_chat_stream() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=PARAMS_WITH_MAX_TOKENS,
    )
    response = chat.stream("What's the weather in san francisco")
    for chunk in response:
        assert isinstance(chunk.content, str)


def test_chat_stream_with_reasoning_content() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID_REASONING_CONTENT,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={
            "include_reasoning": True,
            "reasoning_effort": "low",
        },
    )
    response = chat.stream("hello")

    reasoning_content = ""

    for chunk in response:
        assert isinstance(chunk.content, str)
        reasoning_content += chunk.additional_kwargs.get("reasoning_content", "")

    assert reasoning_content


async def test_chat_astream() -> None:
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    messages = [
        ("user", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    ]
    num_tokens = 0

    async for chunk in chat.astream(messages):
        assert chunk is not None
        assert isinstance(chunk, AIMessageChunk)
        num_tokens += len(chunk.content)
    assert num_tokens > 0


def test_chat_invoke_with_streaming() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        streaming=True,
        params=PARAMS_WITH_MAX_TOKENS,
    )
    response = chat.invoke("What's the weather in san francisco")
    assert isinstance(response.content, str)


def test_chat_stream_with_param_in_constructor() -> None:
    params = TextChatParameters(max_tokens=10)
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params,
    )
    response = chat.stream("What's the weather in san francisco")
    for chunk in response:
        assert isinstance(chunk.content, str)


def test_chat_stream_with_param_in_method() -> None:
    params = TextChatParameters(max_tokens=10)
    chat = ChatWatsonx(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    response = chat.stream("What's the weather in san francisco", params=params)
    for chunk in response:
        assert isinstance(chunk.content, str)


def test_chain_invoke() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=PARAMS_WITH_MAX_TOKENS,
    )

    system = "You are a helpful assistant."
    human = "{text}"
    prompt = ChatPromptTemplate.from_messages([("system", system), ("human", human)])

    chain = prompt | chat
    response = chain.invoke({"text": "Explain the importance of low latency for LLMs."})

    assert response
    assert response.content


def test_chain_invoke_2() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=PARAMS_WITH_MAX_TOKENS,
    )
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a helpful assistant that "
                "translates {input_language} to {output_language}.",
            ),
            ("human", "{input}"),
        ]
    )
    chain = prompt | chat

    response = chain.invoke(
        {
            "input_language": "English",
            "output_language": "German",
            "input": "I love programming.",
        }
    )
    assert response
    assert response.content


def test_chat_bind_tools() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get weather report for a city",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                },
            },
        },
    ]

    llm_with_tools = chat.bind_tools(tools=tools)

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "get_weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


def test_chat_bind_tools_tool_choice_auto() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get weather report for a city",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                },
            },
        },
    ]

    llm_with_tools = chat.bind_tools(tools=tools, tool_choice="auto")

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "get_weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


@pytest.mark.xfail(reason="Not supported yet")
def test_chat_bind_tools_tool_choice_none() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get weather report for a city",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                },
            },
        },
    ]

    llm_with_tools = chat.bind_tools(tools=tools, tool_choice="none")

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "get_weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


def test_chat_bind_tools_tool_choice_required() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get weather report for a city",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                },
            },
        },
    ]

    llm_with_tools = chat.bind_tools(tools=tools, tool_choice="required")

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "get_weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


def test_chat_bind_tools_tool_choice_as_class() -> None:
    """Test that tool choice is respected."""
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    class Person(BaseModel):
        name: str
        age: int

    with_tool = chat.bind_tools([Person])

    result = with_tool.invoke("Erick, 27 years old")
    assert isinstance(result, AIMessage)
    assert result.content == ""  # should just be tool call
    assert len(result.tool_calls) == 1
    tool_call = result.tool_calls[0]
    assert tool_call["name"] == "Person"
    assert tool_call["args"] == {
        "age": 27,
        "name": "Erick",
    }


def test_chat_bind_tools_tool_choice_as_dict() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    class Person(BaseModel):
        name: str
        age: int

    tool_choice = {"type": "function", "function": {"name": "Person"}}

    with_tool = chat.bind_tools([Person], tool_choice=tool_choice)

    result = with_tool.invoke("Erick, 27 years old. Make sure to use correct name")
    assert isinstance(result, AIMessage)
    assert result.content == ""  # should just be tool call
    tool_call = result.tool_calls[0]
    assert tool_call["name"] == "Person"
    assert tool_call["args"] == {
        "age": 27,
        "name": "Erick",
    }


def test_chat_bind_tools_list_tool_choice_dict() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID_TOOL,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    @tool
    def add(a: int, b: int) -> int:
        """Adds a and b."""
        return a + b

    @tool
    def multiply(a: int, b: int) -> int:
        """Multiplies a and b."""
        return a * b

    @tool
    def get_word_length(word: str) -> int:
        """Get word length."""
        return len(word)

    tools = [add, multiply, get_word_length]

    tool_choice = {
        "type": "function",
        "function": {
            "name": "add",
        },
    }

    chat_with_tools = chat.bind_tools(tools, tool_choice=tool_choice)

    query = "What is 3 + 12? "
    resp = chat_with_tools.invoke(query)

    assert resp.content == ""


def test_chat_bind_tools_list_tool_choice_auto() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID_TOOL_2,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    @tool
    def add(a: int, b: int) -> int:
        """Adds a and b."""
        return a + b

    @tool
    def multiply(a: int, b: int) -> int:
        """Multiplies a and b."""
        return a * b

    @tool
    def get_word_length(word: str) -> int:
        """Get word length."""
        return len(word)

    tools = [add, multiply, get_word_length]
    chat_with_tools = chat.bind_tools(tools, tool_choice="auto")

    query = "What is 3 + 12? "
    resp = chat_with_tools.invoke(query)
    assert resp.content == ""
    assert len(resp.tool_calls) == 1
    tool_call = resp.tool_calls[0]
    assert tool_call["name"] == "add"

    query = "What is 3 * 12? "
    resp = chat_with_tools.invoke(query)
    assert resp.content == ""
    assert len(resp.tool_calls) == 1
    tool_call = resp.tool_calls[0]
    assert tool_call["name"] == "multiply"

    query = "Who was the famous painter from Italy?"
    resp = chat_with_tools.invoke(query)
    assert resp.content
    assert len(resp.tool_calls) == 0


def test_chat_bind_tools_with_watsonx_tools() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    toolkit = WatsonxToolkit(
        url=URL,
    )
    weather_tool = toolkit.get_tool("Weather")

    tools = [weather_tool]

    llm_with_tools = chat.bind_tools(tools=tools)

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "Weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


def test_chat_bind_tools_with_watsonx_tools_tool_choice_auto() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    toolkit = WatsonxToolkit(
        url=URL,
    )
    weather_tool = toolkit.get_tool("Weather")

    tools = [weather_tool]

    llm_with_tools = chat.bind_tools(tools=tools, tool_choice="auto")

    response = llm_with_tools.invoke("what's the weather in san francisco, ca")
    assert isinstance(response, AIMessage)
    assert not response.content
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "Weather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


def test_chat_bind_tools_with_watsonx_tools_tool_choice_as_dict() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID_TOOL,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    toolkit = WatsonxToolkit(
        url=URL,
    )
    weather_tool = toolkit.get_tool("Weather")

    tools = [weather_tool]

    tool_choice = {"type": "function", "function": {"name": "Weather"}}

    with_tool = chat.bind_tools(tools, tool_choice=tool_choice)

    result = with_tool.invoke("What's the weather in Boston?")
    assert isinstance(result, AIMessage)
    assert result.content == ""  # should just be tool call
    tool_call = result.tool_calls[0]
    assert tool_call["name"] == "Weather"
    assert tool_call["args"] == {
        "location": "Boston",
    }


def test_chat_bind_tools_with_watsonx_tools_list_tool_choice_auto() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    toolkit = WatsonxToolkit(
        url=URL,
    )
    weather_tool = toolkit.get_tool("Weather")
    google_search_tool = toolkit.get_tool("GoogleSearch")

    tools = [weather_tool, google_search_tool]
    chat_with_tools = chat.bind_tools(tools, tool_choice="auto")

    query = "What is the weather in Boston?"
    resp = chat_with_tools.invoke(query)
    assert resp.content == ""
    assert len(resp.tool_calls) == 1
    tool_call = resp.tool_calls[0]
    assert tool_call["name"] == "Weather"

    query = "Search for IBM"
    resp = chat_with_tools.invoke(query)
    assert resp.content == ""
    assert len(resp.tool_calls) == 1
    tool_call = resp.tool_calls[0]
    assert tool_call["name"] == "GoogleSearch"

    query = "How are you doing?"
    resp = chat_with_tools.invoke(query)
    assert resp.content
    assert len(resp.tool_calls) == 0


def test_chat_bind_tools_with_watsonx_tools_list_tool_choice_dict() -> None:
    """Test that tool choice is respected just passing in True."""
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params={"temperature": 0},
    )

    toolkit = WatsonxToolkit(
        url=URL,
    )
    weather_tool = toolkit.get_tool("Weather")
    google_search_tool = toolkit.get_tool("GoogleSearch")

    tools = [weather_tool, google_search_tool]

    tool_choice = {
        "type": "function",
        "function": {
            "name": "GoogleSearch",
        },
    }

    chat_with_tools = chat.bind_tools(tools, tool_choice=tool_choice)

    query = "What is the weather in Boston?"
    resp = chat_with_tools.invoke(query)
    assert resp.content == ""
    assert len(resp.tool_calls) == 1
    tool_call = resp.tool_calls[0]
    assert tool_call["name"] == "GoogleSearch"


def test_chat_with_json_mode() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    response = chat.invoke(
        "Return this as json: {'a': 1}",
        params={"response_format": {"type": "json_object"}},
    )
    assert isinstance(response.content, str)
    assert json.loads(response.content) == {"a": 1}

    # Test streaming
    full: BaseMessageChunk | None = None
    for chunk in chat.stream(
        "Return this as json: {'a': 1}",
        params={"response_format": {"type": "json_object"}},
    ):
        full = chunk if full is None else full + chunk
    assert isinstance(full, AIMessageChunk)
    assert isinstance(full.content, str)
    assert json.loads(full.content) == {"a": 1}


async def test_chat_with_json_mode_async() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    response = await chat.ainvoke(
        "Return this as json: {'a': 1}",
        params={"response_format": {"type": "json_object"}},
    )
    assert isinstance(response.content, str)
    assert json.loads(response.content) == {"a": 1}

    # Test streaming
    full: BaseMessageChunk | None = None
    async for chunk in chat.astream(
        "Return this as json: {'a': 1}",
        params={"response_format": {"type": "json_object"}},
    ):
        full = chunk if full is None else full + chunk
    assert isinstance(full, AIMessageChunk)
    assert isinstance(full.content, str)
    assert json.loads(full.content) == {"a": 1}


@pytest.mark.xfail(reason="Not implemented")
def test_chat_streaming_tool_call() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    class Person(BaseModel):
        name: str
        age: int

    tool_choice = {"type": "function", "function": {"name": "Person"}}

    tool_llm = chat.bind_tools([Person], tool_choice=tool_choice)

    stream_response = tool_llm.stream("Erick, 27 years old")

    additional_kwargs = None
    for chunk in stream_response:
        assert isinstance(chunk, AIMessageChunk)
        assert chunk.content == ""
        additional_kwargs = chunk.additional_kwargs

    assert additional_kwargs is not None
    assert "tool_calls" in additional_kwargs
    assert len(additional_kwargs["tool_calls"]) == 1
    assert additional_kwargs["tool_calls"][0]["function"]["name"] == "Person"
    assert json.loads(additional_kwargs["tool_calls"][0]["function"]["arguments"]) == {
        "name": "Erick",
        "age": 27,
    }

    assert isinstance(chunk, AIMessageChunk)
    assert len(chunk.tool_call_chunks) == 1
    tool_call_chunk = chunk.tool_call_chunks[0]
    assert tool_call_chunk["name"] == "Person"
    assert tool_call_chunk["args"] == '{"name": "Erick", "age": 27}'

    # where it doesn't call the tool
    strm = tool_llm.stream("What is 2+2?")
    acc: Any = None
    for chunk in strm:
        assert isinstance(chunk, AIMessageChunk)
        acc = chunk if acc is None else acc + chunk
    assert acc.content != ""
    assert "tool_calls" not in acc.additional_kwargs


def test_chat_streaming_multiple_tool_call() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        temperature=0,
    )

    @tool("search")
    def search(query: str) -> list[str]:  # noqa: ARG001
        """Call to search the web for capital of countries"""
        return ["capital of america is washington D.C."]

    @tool("get_weather")
    def get_weather(city: Literal["nyc"]) -> str:
        """Use this to get weather information."""
        if city == "nyc":
            return "It might be cloudy in nyc"
        error_msg = "Unknown city"  # type: ignore[unreachable]
        raise ValueError(error_msg)

    tools = [search, get_weather]
    tools_name = {el.name for el in tools}

    tool_llm = chat.bind_tools(tools)

    stream_response = tool_llm.stream(
        "What is the weather in the NY and what is capital of USA?"
    )

    ai_message = None

    for chunk in stream_response:
        if ai_message is None:
            ai_message = chunk
        else:
            ai_message += chunk  # type: ignore[assignment]
        print(chunk.id, type(chunk.id))
        assert isinstance(chunk, AIMessageChunk)
        assert chunk.content == ""

    ai_message = cast("AIMessageChunk", ai_message)
    assert ai_message.response_metadata.get("finish_reason") == "tool_calls"
    assert ai_message.response_metadata.get("model_name") == MODEL_ID
    assert ai_message.id is not None

    # additional_kwargs
    assert ai_message.additional_kwargs is not None
    assert "tool_calls" in ai_message.additional_kwargs
    assert len(ai_message.additional_kwargs["tool_calls"]) == 2
    assert {
        el["function"]["name"] for el in ai_message.additional_kwargs["tool_calls"]
    } == tools_name

    # tool_calls
    assert all(el["id"] is not None for el in ai_message.tool_calls)
    assert all(el["type"] == "tool_call" for el in ai_message.tool_calls)
    assert {el["name"] for el in ai_message.tool_calls} == tools_name

    generated_tools_args = [{"city": "nyc"}, {"query": "capital of USA"}]
    assert {next(iter(el["args"].keys())) for el in ai_message.tool_calls} == {
        next(iter(el.keys())) for el in generated_tools_args
    }

    # tool_call_chunks
    predicted_tool_call_chunks = []
    for i, el in enumerate(ai_message.tool_calls):
        tool_call_chunk = ToolCallChunk(
            name=el["name"],
            args=json.dumps(el["args"]),
            id=el["id"],
            index=i,
            type="tool_call_chunk",
        )
        predicted_tool_call_chunks.append(tool_call_chunk)

    assert ai_message.tool_call_chunks == predicted_tool_call_chunks
    assert (
        json.loads(
            ai_message.additional_kwargs["tool_calls"][0]["function"]["arguments"]
        )
        == generated_tools_args[0]
    )
    assert (
        json.loads(
            ai_message.additional_kwargs["tool_calls"][1]["function"]["arguments"]
        )
        == generated_tools_args[1]
    )

    assert ai_message.usage_metadata is not None


def test_chat_structured_output_function_calling() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        temperature=0,
    )
    schema = {
        "title": "AnswerWithJustification",
        "description": (
            "An answer to the user question along with justification for the answer."
        ),
        "type": "object",
        "properties": {
            "answer": {"title": "Answer", "type": "string"},
            "justification": {"title": "Justification", "type": "string"},
        },
        "required": ["answer", "justification"],
    }
    structured_llm = chat.with_structured_output(schema, method="function_calling")
    result = structured_llm.invoke(
        "What weighs more a pound of bricks or a pound of feathers"
    )
    assert isinstance(result, dict)
    assert "answer" in result
    assert "justification" in result


def test_chat_structured_output_json_schema() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        temperature=0,
    )
    schema = {
        "title": "AnswerWithJustification",
        "description": (
            "An answer to the user question along with justification for the answer."
        ),
        "type": "object",
        "properties": {
            "answer": {"title": "Answer", "type": "string"},
            "justification": {"title": "Justification", "type": "string"},
        },
        "required": ["answer", "justification"],
    }
    structured_llm = chat.with_structured_output(schema, method="json_schema")
    result = structured_llm.invoke(
        "What weighs more a pound of bricks or a pound of feathers"
    )
    assert isinstance(result, dict)
    assert "answer" in result
    assert "justification" in result


def test_chat_streaming_structured_output_function_calling() -> None:
    chat = ChatWatsonx(
        model_id=MODEL_ID_TOOL,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    class Person(BaseModel):
        name: str
        age: int

    structured_llm = chat.with_structured_output(Person)
    stream_response = structured_llm.stream("Erick, 123456 years old")

    for chunk in stream_response:
        assert isinstance(chunk, Person)
        assert chunk.name in "Erick"
        assert chunk.age == 123456


###################################
# --------Parameter-Tests-------- #
###################################

prompt_1 = "Say: 'Hello, My name is Erick!'"


def test_init_with_params_1() -> None:
    params_1 = None
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1,
    )
    assert chat.params == {}


def test_init_with_params_2() -> None:
    params_1 = {"max_tokens": 10}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1,
    )
    assert chat.params == params_1


def test_init_with_params_3() -> None:
    params_1 = {"max_tokens": 10}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        **params_1,
    )
    assert chat.params == params_1


def test_init_with_params_4() -> None:
    params_1 = {"max_tokens": 10}
    params_2 = {"temperature": 0.5}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1,
        **params_2,
    )
    assert chat.params == params_1 | params_2


def test_init_with_params_5() -> None:
    params_1 = {"max_tokens": 10}
    params_2 = {"max_tokens": 20}

    pattern = re.escape(
        "Duplicate parameters found in params and keyword arguments: ['max_tokens']"
    )

    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            url=URL,
            project_id=WX_PROJECT_ID,
            params=params_1,
            **params_2,
        )


def test_invoke_with_params_1() -> None:
    params_1 = None
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    resp = chat.invoke(prompt_1, params=params_1)
    completion_tokens = resp.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )

    assert chat.params == {}
    assert completion_tokens > 0


def test_invoke_with_params_2() -> None:
    params_1 = {"max_tokens": 5}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    resp = chat.invoke(prompt_1, params=params_1)
    completion_tokens = resp.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )

    assert chat.params == {}
    assert completion_tokens == 5


def test_invoke_with_params_3() -> None:
    params_1_a = {"max_tokens": 5}
    params_1_b = {"max_tokens": 10}
    params_2_a = {"logprobs": False}
    params_2_b = {"logprobs": True}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    resp_1 = chat.invoke(prompt_1, params=params_1_a, **params_2_a)  # type: ignore[arg-type]
    completion_tokens = resp_1.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    logprobs = resp_1.response_metadata.get("logprobs")

    assert chat.params == {}
    assert completion_tokens == 5
    assert not logprobs

    resp_2 = chat.invoke(prompt_1, params=params_1_a, **params_2_b)  # type: ignore[arg-type]
    completion_tokens = resp_2.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    logprobs = resp_2.response_metadata.get("logprobs")

    assert chat.params == {}
    assert completion_tokens == 5
    assert logprobs

    resp_3 = chat.invoke(prompt_1, **params_1_b, **params_2_b)  # type: ignore[arg-type]
    completion_tokens = resp_3.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    logprobs = resp_3.response_metadata.get("logprobs")

    assert chat.params == {}
    assert 7 < completion_tokens < 11
    assert logprobs


def test_invoke_with_params_4() -> None:
    params_1 = {"max_tokens": 5}
    params_2 = {"max_tokens": 20}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    pattern = re.escape(
        "Duplicate parameters found in params and keyword arguments: ['max_tokens']"
    )
    with pytest.raises(ValueError, match=pattern):
        chat.invoke(prompt_1, params=params_1, **params_2)  # type: ignore[arg-type]


def test_invoke_with_params_5() -> None:
    params_1 = {"max_tokens": 5, "logprobs": False}
    params_2 = {"max_tokens": 10}
    params_3 = {"logprobs": True}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    pattern = (
        r"(?=.*Duplicate parameters found in params and keyword arguments: )"
        r"(?=.*'logprobs')(?=.*'max_tokens')"
    )
    with pytest.raises(ValueError, match=pattern):
        chat.invoke(prompt_1, params=params_1, **params_2, **params_3)  # type: ignore[arg-type]


def test_init_and_invoke_with_params_1() -> None:
    params_1 = {"max_tokens": 11}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1,
    )
    resp = chat.invoke(prompt_1, params=params_1)
    completion_tokens = resp.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    assert chat.params == params_1
    assert 7 < completion_tokens <= 11


def test_init_and_invoke_with_params_2() -> None:
    params_1_a = {"max_tokens": 4}
    params_1_b = {"max_tokens": 5}
    params_1_c = {"max_tokens": 6}

    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1_a,
    )
    resp_1 = chat.invoke(prompt_1, params=params_1_b)
    completion_tokens = resp_1.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    assert chat.params == params_1_a
    assert completion_tokens == 5

    resp_2 = chat.invoke(prompt_1, **params_1_c)  # type: ignore[arg-type]
    completion_tokens = resp_2.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    assert chat.params == params_1_a
    assert completion_tokens == 6

    resp_2 = chat.invoke(prompt_1)
    completion_tokens = resp_2.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    assert chat.params == params_1_a
    assert completion_tokens == 4


def test_init_and_invoke_with_params_3() -> None:
    params_1_a = {"max_tokens": 2}
    params_1_b = {"max_tokens": 5}
    params_2_a = {"logprobs": False}
    params_2_b = {"logprobs": True}
    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1_a,
        **params_2_a,
    )
    resp_1 = chat.invoke(prompt_1, params=params_1_b, **params_2_b)  # type: ignore[arg-type]
    completion_tokens = resp_1.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    logprobs = resp_1.response_metadata.get("logprobs")

    assert chat.params == params_1_a | params_2_a
    assert completion_tokens == 5
    assert logprobs

    resp_2 = chat.invoke(prompt_1)
    completion_tokens = resp_2.response_metadata.get("token_usage", {}).get(
        "completion_tokens"
    )
    logprobs = resp_2.response_metadata.get("logprobs")

    assert chat.params == params_1_a | params_2_a
    assert completion_tokens == 2
    assert not logprobs


def test_init_and_invoke_with_params_4() -> None:
    params_1_a = {"max_tokens": 4}
    params_1_b = {"max_tokens": 5}
    params_1_c = {"max_tokens": 6}

    chat = ChatWatsonx(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1_a,
    )

    pattern = re.escape(
        "Duplicate parameters found in params and keyword arguments: ['max_tokens']"
    )

    with pytest.raises(ValueError, match=pattern) as e:
        chat.invoke(prompt_1, params=params_1_b, **params_1_c)  # type: ignore[arg-type]

    assert (
        "Duplicate parameters found in params and keyword arguments: ['max_tokens']"
        in str(e.value)
    )



import os
from collections.abc import Generator
from typing import Any
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd  # type: ignore[import-untyped]
import pytest
from pyarrow import flight  # type: ignore[import-untyped]

from langchain_ibm.utilities.sql_database import (
    WatsonxSQLDatabase,
    pretty_print_table_info,
    truncate_word,
)

CONNECTION_ID = "test_connection_id"
PROJECT_ID = "test_project_id"


@pytest.fixture
def schema() -> str:
    return "test_schema"


@pytest.fixture
def table_name() -> str:
    return "test_table"


@pytest.fixture
def table_info() -> dict[str, Any]:
    return {
        "fields": [
            {"name": "id", "type": {"native_type": "INT", "nullable": False}},
            {"name": "name", "type": {"native_type": "VARCHAR(255)", "nullable": True}},
            {"name": "age", "type": {"native_type": "INT", "nullable": True}},
        ],
        "extended_metadata": [
            {"name": "primary_key", "value": {"key_columns": ["id"]}}
        ],
    }


@pytest.fixture
def clear_env() -> Generator[None, None, None]:
    with mock.patch.dict(os.environ, clear=True):
        yield


class MockFlightSQLClient:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __enter__(self, *args: Any, **kwargs: Any) -> "MockFlightSQLClient":
        return self

    def __exit__(self, *args: object) -> None:
        pass

    def get_tables(self, *_args: Any, **_kwargs: Any) -> dict[str, Any]:
        return {"assets": [{"name": "table1"}, {"name": "table2"}]}

    def get_table_info(
        self, table_name: str, *_args: Any, **_kwargs: Any
    ) -> dict[str, Any]:
        if table_name == "table1":
            return {
                "path": "/public/table1",
                "fields": [
                    {"name": "id", "type": {"native_type": "INT", "nullable": False}},
                    {
                        "name": "name",
                        "type": {"native_type": "VARCHAR(255)", "nullable": True},
                    },
                    {"name": "age", "type": {"native_type": "INT", "nullable": True}},
                ],
                "extended_metadata": [
                    {"name": "primary_key", "value": {"key_columns": ["id"]}}
                ],
            }
        if table_name == "table2":
            return {
                "path": "/public/table2",
                "fields": [
                    {"name": "id", "type": {"native_type": "INT", "nullable": False}},
                    {
                        "name": "name",
                        "type": {"native_type": "VARCHAR(255)", "nullable": True},
                    },
                    {"name": "age", "type": {"native_type": "INT", "nullable": True}},
                ],
                "extended_metadata": [
                    {"name": "primary_key", "value": {"key_columns": ["id"]}}
                ],
            }
        error_msg = "Table not found"
        raise flight.FlightError(error_msg)

    def execute(self, *_args: Any, **kwargs: Any) -> pd.DataFrame:
        if "table1" in kwargs.get("query", ""):
            return pd.DataFrame({"id": [1], "name": ["test"], "age": [35]})

        if "table1" not in kwargs.get("query", ""):
            error_msg = "Table not found"
            raise flight.FlightError(error_msg)

        error_msg = "syntax error"
        raise ValueError(error_msg)

    def get_n_first_rows(self, *_args: Any, **_kwargs: Any) -> pd.DataFrame:
        return pd.DataFrame({"id": [1], "name": ["test"], "age": [35]})


### truncate_word


def test_truncate_word() -> None:
    assert truncate_word("This is a test", length=11) == "This is..."
    assert truncate_word("Short", length=10) == "Short"
    assert truncate_word("A", length=3) == "A"
    assert truncate_word("", length=10) == ""
    assert (
        truncate_word("This is a longer test", length=20, suffix="***")
        == "This is a longer***"
    )
    assert truncate_word(12345, length=10) == 12345  # Non-string input
    assert truncate_word("This is a test", length=0) == "This is a test"  # Length <= 0


def test_truncate_word_edge_cases() -> None:
    assert truncate_word("This is a test", length=8) == "This..."
    assert truncate_word("This is a test", length=11) == "This is..."
    assert truncate_word("This is a test", length=13) == "This is a..."
    assert (
        truncate_word("This is a test", length=16) == "This is a test"
    )  # Length >= string lengths


### pretty_print_table_info


def test_pretty_print_table_info(
    schema: str, table_name: str, table_info: dict[str, Any]
) -> None:
    expected_output = """
CREATE TABLE "test_schema"."test_table" (
\t"id" INT NOT NULL,
\t"name" VARCHAR(255),
\t"age" INT,
\tCONSTRAINT primary_key PRIMARY KEY (id)
\t)"""
    assert pretty_print_table_info(schema, table_name, table_info) == expected_output


def test_pretty_print_table_info_with_nullable_columns() -> None:
    schema = "another_schema"
    table_name = "another_table"
    table_info = {
        "fields": [
            {
                "name": "email",
                "type": {"native_type": "VARCHAR(255)", "nullable": True},
            },
            {
                "name": "created_at",
                "type": {"native_type": "TIMESTAMP", "nullable": True},
            },
        ],
        "extended_metadata": [
            {"name": "primary_key", "value": {"key_columns": ["email"]}}
        ],
    }
    expected_output = """
CREATE TABLE "another_schema"."another_table" (
\t"email" VARCHAR(255),
\t"created_at" TIMESTAMP,
\tCONSTRAINT primary_key PRIMARY KEY (email)
\t)"""
    assert pretty_print_table_info(schema, table_name, table_info) == expected_output


def test_pretty_print_table_info_without_primary_key() -> None:
    schema = "no_pk_schema"
    table_name = "no_pk_table"
    table_info = {
        "fields": [
            {"name": "value1", "type": {"native_type": "INT", "nullable": False}},
            {
                "name": "value2",
                "type": {"native_type": "VARCHAR(255)", "nullable": True},
            },
        ]
    }
    expected_output = """
CREATE TABLE "no_pk_schema"."no_pk_table" (
\t"value1" INT NOT NULL,
\t"value2" VARCHAR(255)
\t)"""
    assert pretty_print_table_info(schema, table_name, table_info) == expected_output


### WatsonSQLDatabase


def test_initialize_watsonx_sql_database_without_url(schema: str) -> None:
    pattern = r"(?=.*url)(?=.*WATSONX_URL)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)


def test_initialize_watsonx_sql_database_cloud_only_url(schema: str) -> None:
    pattern = (
        r"(?=.*api_key)(?=.*token)(?!.*password)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_TOKEN)(?!.*WATSONX_PASSWORD)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(
            connection_id=CONNECTION_ID,
            schema=schema,
            url="https://us-south.ml.cloud.ibm.com",
        )


def test_initialize_watsonx_sql_database_cpd_bad_path_without_all(schema: str) -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(
            connection_id=CONNECTION_ID,
            schema=schema,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
        )


def test_initialize_watsonx_sql_database_cpd_bad_path_password_without_username(
    schema: str,
) -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(
            connection_id=CONNECTION_ID,
            schema=schema,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            password="test_password",  # noqa: S106
        )


def test_initialize_watsonx_sql_database_cpd_bad_path_apikey_without_username(
    schema: str,
) -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(
            connection_id=CONNECTION_ID,
            schema=schema,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_sql_database_cpd_bad_path_without_instance_id(
    schema: str,
) -> None:
    pattern = r"(?=.*instance_id)(?=.*WATSONX_INSTANCE_ID)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxSQLDatabase(
            connection_id=CONNECTION_ID,
            schema=schema,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
            username="test_user",
        )


def test_initialize_watsonx_sql_database_without_any_params() -> None:
    with pytest.raises(TypeError):
        WatsonxSQLDatabase()  # type: ignore[call-arg]


def test_initialize_watsonx_sql_database_valid(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)

        assert isinstance(wx_sql_database._flight_sql_client, MockFlightSQLClient)
        assert wx_sql_database.schema == schema


def test_initialize_watsonx_sql_database_include_tables(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(
            connection_id=CONNECTION_ID, schema=schema, include_tables=["table1"]
        )

        assert wx_sql_database.get_usable_table_names() == ["table1"]


def test_initialize_watsonx_sql_database_ignore_tables(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(
            connection_id=CONNECTION_ID, schema=schema, ignore_tables=["table1"]
        )

        assert wx_sql_database.get_usable_table_names() == ["table2"]


def test_initialize_watsonx_sql_database_get_table_info(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)
        expected_output = """
CREATE TABLE "test_schema"."table1" (
\t"id" INT NOT NULL,
\t"name" VARCHAR(255),
\t"age" INT,
\tCONSTRAINT primary_key PRIMARY KEY (id)
\t)

First 3 rows of table table1:

 id name  age
  1 test   35"""
        print(wx_sql_database.get_table_info(["table1"]))
        assert wx_sql_database.get_table_info(["table1"]) == expected_output


def test_initialize_watsonx_sql_database_get_table_info_no_throw(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)
        with pytest.raises(
            ValueError, match=r"table_names {'tableX'} not found in database"
        ):
            wx_sql_database.get_table_info(["tableX"])


def test_initialize_watsonx_sql_database_run(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)

        assert (
            wx_sql_database.run(f"SELECT * FROM {schema}.table1", include_columns=True)  # noqa: S608
            == "[{'id': 1, 'name': 'test', 'age': 35}]"
        )


def test_initialize_watsonx_sql_database_run_no_throw(
    schema: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_api_client = Mock()
    mock_api_client.default_project_id = PROJECT_ID

    with (
        mock.patch.dict(os.environ, clear=True),
        patch(
            "langchain_ibm.utilities.sql_database.APIClient",
            autospec=True,
            return_value=mock_api_client,
        ),
        patch(
            "langchain_ibm.utilities.sql_database.FlightSQLClient",
            autospec=True,
            return_value=MockFlightSQLClient(),
        ),
    ):
        envvars = {
            "WATSONX_APIKEY": "test_apikey",
            "WATSONX_URL": "https://us-south.ml.cloud.ibm.com",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)

        wx_sql_database = WatsonxSQLDatabase(connection_id=CONNECTION_ID, schema=schema)

        assert "Table not found" in wx_sql_database.run_no_throw(
            f"SELECT * FROM {schema}.tableX",  # noqa: S608
            include_columns=True,
        )



"""IBM watsonx.ai rerank wrapper."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import Rerank  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    RerankParameters,
)
from langchain_core.documents import BaseDocumentCompressor, Document
from langchain_core.utils.utils import secret_from_env
from pydantic import AliasChoices, ConfigDict, Field, SecretStr, model_validator
from typing_extensions import Self, override

from langchain_ibm.utils import (
    extract_params,
    normalize_api_key,
    resolve_watsonx_credentials,
    secret_from_env_multi,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from langchain_core.callbacks import Callbacks


class WatsonxRerank(BaseDocumentCompressor):
    """Document compressor that uses `watsonx Rerank API`.

    ???+ info "Setup"

        To use, you should have `langchain_ibm` python package installed,
        and the environment variable `WATSONX_API_KEY` set with your API key, or pass
        it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        ```python
        from langchain_ibm import WatsonxRerank
        from ibm_watsonx_ai.foundation_models.schema import RerankParameters

        parameters = RerankParameters(truncate_input_tokens=20)

        ranker = WatsonxRerank(
            model_id="cross-encoder/ms-marco-minilm-l-12-v2",
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",
            params=parameters,
            # api_key="*****"
        )
        ```

    ??? info "Rerank"

        ```python
        query = "red cat chasing a laser pointer"
        documents = [
            "A red cat darts across the living room, pouncing on a red laser dot.",
            "Two dogs play fetch in the park with a tennis ball.",
            "The tabby cat naps on a sunny windowsill all afternoon.",
            "A recipe for tuna casserole with crispy breadcrumbs.",
        ]

        ranker.rerank(documents=documents, query=query)
        ```

        ```python
        [
            {"index": 0, "relevance_score": 0.8719543218612671},
            {"index": 2, "relevance_score": 0.6520894169807434},
            {"index": 1, "relevance_score": 0.6270776391029358},
            {"index": 3, "relevance_score": 0.4607713520526886},
        ]
        ```

    """

    model_id: str
    """Type of model to use."""

    project_id: str | None = None
    """ID of the Watson Studio project."""

    space_id: str | None = None
    """ID of the Watson Studio space."""

    url: SecretStr = Field(
        alias="url",
        default_factory=secret_from_env("WATSONX_URL", default=None),  # type: ignore[assignment]
    )
    """URL to the Watson Machine Learning or CPD instance."""

    apikey: SecretStr | None = None
    api_key: SecretStr | None = Field(
        default_factory=secret_from_env_multi(
            names_priority=["WATSONX_API_KEY", "WATSONX_APIKEY"],
            deprecated={"WATSONX_APIKEY"},
        ),
        serialization_alias="api_key",
        validation_alias=AliasChoices("api_key", "apikey"),  # accept both on input
        description="API key to the Watson Machine Learning or CPD instance.",
    )
    """API key to the Watson Machine Learning or CPD instance."""

    token: SecretStr | None = Field(
        alias="token",
        default_factory=secret_from_env("WATSONX_TOKEN", default=None),
    )
    """Token to the CPD instance."""

    password: SecretStr | None = Field(
        alias="password",
        default_factory=secret_from_env("WATSONX_PASSWORD", default=None),
    )
    """Password to the CPD instance."""

    username: SecretStr | None = Field(
        alias="username",
        default_factory=secret_from_env("WATSONX_USERNAME", default=None),
    )
    """Username to the CPD instance."""

    instance_id: SecretStr | None = Field(
        alias="instance_id",
        default_factory=secret_from_env("WATSONX_INSTANCE_ID", default=None),
    )
    """Instance_id of the CPD instance."""

    version: SecretStr | None = None
    """Version of the CPD instance."""

    params: dict[str, Any] | RerankParameters | None = None
    """Model parameters to use during request generation."""

    verify: str | bool | None = None
    """You can pass one of following as verify:
        * the path to a CA_BUNDLE file
        * the path of directory with certificates of trusted CAs
        * True - default path to truststore will be taken
        * False - no verification will be made"""

    validate_model: bool = True
    """Model ID validation."""

    streaming: bool = False
    """ Whether to stream the results or not. """

    watsonx_rerank: Rerank = Field(default=None, exclude=True)  #: :meta private:

    watsonx_client: APIClient | None = Field(default=None, exclude=True)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        protected_namespaces=(),
    )

    @property
    def lc_secrets(self) -> dict[str, str]:
        """Mapping of secret environment variables."""
        return {
            "url": "WATSONX_URL",
            "api_key": "WATSONX_API_KEY",  # preferred
            "apikey": "WATSONX_APIKEY",
            "token": "WATSONX_TOKEN",
            "password": "WATSONX_PASSWORD",
            "username": "WATSONX_USERNAME",
            "instance_id": "WATSONX_INSTANCE_ID",
        }

    @model_validator(mode="before")
    @classmethod
    def _normalize_and_warn_deprecated_input(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Handle deprecated input kwarg name `apikey` vs new `api_key`.
            data = normalize_api_key(data=data)
        return data

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that credentials and python package exists in environment."""
        if isinstance(self.watsonx_client, APIClient):
            watsonx_rerank = Rerank(
                model_id=self.model_id,
                params=self.params,
                api_client=self.watsonx_client,
                project_id=self.project_id,
                space_id=self.space_id,
                verify=self.verify,
            )
            self.watsonx_rerank = watsonx_rerank

        else:
            credentials = resolve_watsonx_credentials(
                url=self.url,
                api_key=self.api_key,
                token=self.token,
                password=self.password,
                username=self.username,
                instance_id=self.instance_id,
                version=self.version,
                verify=self.verify,
            )

            watsonx_rerank = Rerank(
                model_id=self.model_id,
                credentials=credentials,
                params=self.params,
                project_id=self.project_id,
                space_id=self.space_id,
                verify=self.verify,
            )
            self.watsonx_rerank = watsonx_rerank

        return self

    def rerank(
        self,
        documents: Sequence[str | Document | dict[str, Any]],
        query: str,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Rerank documents."""
        if len(documents) == 0:  # to avoid empty api call
            return []
        docs = [
            doc.page_content if isinstance(doc, Document) else doc for doc in documents
        ]
        params = extract_params(kwargs, self.params)

        results = self.watsonx_rerank.generate(
            query=query,
            inputs=docs,
            **(kwargs | {"params": params}),
        )
        return [
            {"index": res.get("index"), "relevance_score": res.get("score")}
            for res in results["results"]
        ]

    @override
    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Callbacks | None = None,
        **kwargs: Any,
    ) -> Sequence[Document]:
        """Compress documents using watsonx's rerank API.

        Args:
            documents: A sequence of documents to compress
            query: The query to use for compressing the documents
            callbacks: Callbacks to run during the compression process
            kwargs: Additional keyword args

        Returns:
            A sequence of compressed documents
        """
        compressed = []
        for res in self.rerank(documents, query, **kwargs):
            doc = documents[res["index"]]
            doc_copy = Document(doc.page_content, metadata=deepcopy(doc.metadata))
            doc_copy.metadata["relevance_score"] = res["relevance_score"]
            compressed.append(doc_copy)
        return compressed



"""Test WatsonxLLM API wrapper.

You'll need to set WATSONX_APIKEY and WATSONX_PROJECT_ID environment variables.
"""

import os
from typing import Any

import pytest
from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import (  # type: ignore[import-untyped]
    Model,
    ModelInference,
)
from ibm_watsonx_ai.foundation_models.utils.enums import (  # type: ignore[import-untyped]
    DecodingMethods,
)
from ibm_watsonx_ai.metanames import (  # type: ignore[import-untyped]
    GenTextParamsMetaNames,
)
from langchain_core.outputs import LLMResult

from langchain_ibm import WatsonxLLM

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

MODEL_ID = "ibm/granite-3-3-8b-instruct"

CREATE_WATSONX_LLM_INIT_PARAMETERS = [
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only api_key",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only apikey",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="api_key and apikey",
    ),
]


@pytest.mark.parametrize("init_data", CREATE_WATSONX_LLM_INIT_PARAMETERS)
def test_watsonxllm_init(init_data: dict[str, Any]) -> None:
    watsonxllm = WatsonxLLM(**init_data)

    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_with_params() -> None:
    parameters = {
        GenTextParamsMetaNames.TEMPERATURE: 0,
        GenTextParamsMetaNames.STOP_SEQUENCES: ["am"],
    }

    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters,
    )
    response = watsonxllm.invoke("Write: 'I am superhero!'\n")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert response.endswith("am")


def test_watsonxllm_invoke_with_params_2() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }

    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.invoke("What color sunflower is?", params=parameters)
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_with_params_3() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
    }
    parameters_2 = {
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }

    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    response = watsonxllm.invoke("What color sunflower is?", params=parameters_2)
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_with_params_4() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
    }
    parameters_2 = {
        "temperature": 0.6,
    }

    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    response = watsonxllm.invoke("What color sunflower is?", **parameters_2)  # type: ignore[arg-type]
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_with_params_5_diff() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.TEMPERATURE: 0,
        GenTextParamsMetaNames.STOP_SEQUENCES: ["am"],
    }
    parameters_2 = {
        GenTextParamsMetaNames.TEMPERATURE: 0,
        GenTextParamsMetaNames.STOP_SEQUENCES: ["random_lorem"],
    }

    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    response_1 = watsonxllm.invoke("Write: 'I am superhero!'\n")
    print(f"\nResponse 1: {response_1}")
    assert isinstance(response_1, str)
    assert response_1.endswith("am")
    response_2 = watsonxllm.invoke("Write: 'I am superhero!'\n", params=parameters_2)
    print(f"\nResponse 2: {response_2}")
    assert isinstance(response_2, str)
    assert not response_2.endswith("am")


def test_watsonxllm_generate() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.generate(["What color sunflower is?"])
    print(f"\nResponse: {response}")
    response_text = response.generations[0][0].text
    print(f"Response text: {response_text}")
    assert isinstance(response, LLMResult)
    assert len(response_text) > 0


def test_watsonxllm_generate_with_param() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.generate(["What color sunflower is?"], params=parameters)
    print(f"\nResponse: {response}")
    response_text = response.generations[0][0].text
    print(f"Response text: {response_text}")
    assert isinstance(response, LLMResult)
    assert len(response_text) > 0


def test_watsonxllm_generate_with_multiple_prompts() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.generate(
        ["What color sunflower is?", "What color turtle is?"]
    )
    print(f"\nResponse: {response}")
    response_text = response.generations[0][0].text
    print(f"Response text: {response_text}")
    assert isinstance(response, LLMResult)
    assert len(response_text) > 0


def test_watsonxllm_invoke_with_guardrails() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.invoke("What color sunflower is?", guardrails=True)
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_with_streaming() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        streaming=True,
    )
    response = watsonxllm.invoke("What color sunflower is?")
    assert isinstance(response, str)


def test_watsonxllm_generate_stream() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.generate(["What color sunflower is?"], stream=True)
    print(f"\nResponse: {response}")
    response_text = response.generations[0][0].text
    print(f"Response text: {response_text}")
    assert isinstance(response, LLMResult)
    assert len(response_text) > 0


def test_watsonxllm_stream() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    stream_response = watsonxllm.stream("What color sunflower is?")

    linked_text_stream = ""
    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        linked_text_stream += chunk
    print(f"Linked text stream: {linked_text_stream}")
    assert linked_text_stream


async def test_watsonxllm_astream() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )

    stream_response = watsonxllm.astream("What color sunflower is?")

    linked_text_stream = ""
    async for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        linked_text_stream += chunk

    assert len(linked_text_stream) > 0


def test_watsonxllm_stream_with_kwargs() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    stream_response = watsonxllm.stream("What color sunflower is?", raw_response=True)

    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )


def test_watsonxllm_stream_with_params() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: "greedy",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters,
    )
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")

    stream_response = watsonxllm.stream("What color sunflower is?")

    linked_text_stream = ""
    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        linked_text_stream += chunk
    print(f"Linked text stream: {linked_text_stream}")
    assert response == linked_text_stream, (
        "Linked text stream are not the same as generated text"
    )


def test_watsonxllm_stream_with_params_2() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: "greedy",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    stream_response = watsonxllm.stream("What color sunflower is?", params=parameters)

    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        print(chunk)


def test_watsonxllm_stream_with_params_3() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
    }
    parameters_2 = {
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 5,
    }
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    stream_response = watsonxllm.stream("What color sunflower is?", params=parameters_2)

    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        print(chunk)


def test_watsonxllm_stream_with_params_4() -> None:
    parameters_1 = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 10,
    }
    parameters_2 = {
        "temperature": 0.6,
    }
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
        params=parameters_1,
    )
    stream_response = watsonxllm.stream("What color sunflower is?", **parameters_2)  # type: ignore[arg-type]

    for chunk in stream_response:
        assert isinstance(chunk, str), (
            f"chunk expect type '{str}', actual '{type(chunk)}'"
        )
        print(chunk)


def test_watsonxllm_invoke_from_wx_model() -> None:
    model = Model(
        model_id=MODEL_ID,
        credentials={
            "apikey": WX_APIKEY,
            "url": "https://us-south.ml.cloud.ibm.com",
        },
        project_id=WX_PROJECT_ID,
    )
    watsonxllm = WatsonxLLM(watsonx_model=model)
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_from_wx_model_inference() -> None:
    credentials = Credentials(
        api_key=WX_APIKEY, url="https://us-south.ml.cloud.ibm.com"
    )
    model = ModelInference(
        model_id=MODEL_ID,
        credentials=credentials,
        project_id=WX_PROJECT_ID,
    )
    watsonxllm = WatsonxLLM(watsonx_model=model)
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_from_wx_model_inference_with_params() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: "sample",
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 100,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 10,
        GenTextParamsMetaNames.TEMPERATURE: 0.5,
        GenTextParamsMetaNames.TOP_K: 50,
        GenTextParamsMetaNames.TOP_P: 1,
    }
    model = ModelInference(
        model_id=MODEL_ID,
        credentials={
            "apikey": WX_APIKEY,
            "url": "https://us-south.ml.cloud.ibm.com",
        },
        project_id=WX_PROJECT_ID,
        params=parameters,
    )
    watsonxllm = WatsonxLLM(watsonx_model=model)
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_watsonxllm_invoke_from_wx_model_inference_with_params_as_enum() -> None:
    parameters = {
        GenTextParamsMetaNames.DECODING_METHOD: DecodingMethods.GREEDY,
        GenTextParamsMetaNames.MAX_NEW_TOKENS: 100,
        GenTextParamsMetaNames.MIN_NEW_TOKENS: 10,
        GenTextParamsMetaNames.TEMPERATURE: 0.5,
        GenTextParamsMetaNames.TOP_K: 50,
        GenTextParamsMetaNames.TOP_P: 1,
    }
    model = ModelInference(
        model_id=MODEL_ID,
        credentials={
            "apikey": WX_APIKEY,
            "url": "https://us-south.ml.cloud.ibm.com",
        },
        project_id=WX_PROJECT_ID,
        params=parameters,
    )
    watsonxllm = WatsonxLLM(watsonx_model=model)
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


async def test_watsonx_ainvoke() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = await watsonxllm.ainvoke("What color sunflower is?")
    assert isinstance(response, str)


async def test_watsonx_acall() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = await watsonxllm._acall("what is the color of the grass?")
    assert "green" in response.lower()


async def test_watsonx_agenerate() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = await watsonxllm.agenerate(
        ["What color sunflower is?", "What color turtle is?"]
    )
    assert len(response.generations) > 0
    assert response.llm_output["token_usage"]["completion_tokens"] != 0  # type: ignore[index]


async def test_watsonx_agenerate_with_stream() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = await watsonxllm.agenerate(["What color sunflower is?"], stream=True)
    assert "yellow" in response.generations[0][0].text.lower()


def test_get_num_tokens() -> None:
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    num_tokens = watsonxllm.get_num_tokens("What color sunflower is?")
    assert num_tokens > 0


def test_init_with_client() -> None:
    watsonx_client = APIClient(
        credentials={
            "url": "https://us-south.ml.cloud.ibm.com",
            "apikey": WX_APIKEY,
        }
    )
    watsonxllm = WatsonxLLM(
        model_id=MODEL_ID,
        watsonx_client=watsonx_client,
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.invoke("What color sunflower is?")
    print(f"\nResponse: {response}")
    assert isinstance(response, str)
    assert len(response) > 0


def test_moderations_generate() -> None:
    guardrails_hap_params = {"input": False, "output": True}
    guardrails_pii_params = {"input": False, "output": True}

    watsonxllm = WatsonxLLM(
        model_id="meta-llama/llama-3-3-70b-instruct",
        url="https://us-south.ml.cloud.ibm.com",
        project_id=WX_PROJECT_ID,
    )
    response = watsonxllm.generate(
        [
            "Please repeat the words in [], do not trim space.\n"
            "[ I hate this damn world. ]"
        ],
        guardrails=True,
        guardrails_pii_params=guardrails_pii_params,
        guardrails_hap_params=guardrails_hap_params,
    )

    assert response
    assert response.generations[0][0].generation_info["moderations"]  # type: ignore[index]



from langchain_ibm import __all__

EXPECTED_ALL = [
    "WatsonxLLM",
    "WatsonxEmbeddings",
    "ChatWatsonx",
    "WatsonxRerank",
]


def test_all_imports() -> None:
    assert sorted(EXPECTED_ALL) == sorted(__all__)



# noqa: D104
import importlib
import warnings
from typing import Any

from langchain_ibm.chat_models import ChatWatsonx
from langchain_ibm.embeddings import WatsonxEmbeddings
from langchain_ibm.llms import WatsonxLLM
from langchain_ibm.rerank import WatsonxRerank

__all__ = [
    "ChatWatsonx",
    "WatsonxEmbeddings",
    "WatsonxLLM",
    "WatsonxRerank",
]


_module_lookup = {
    "WatsonxTool": "langchain_ibm.agent_toolkits.utility.toolkit",
    "WatsonxToolkit": "langchain_ibm.agent_toolkits.utility.toolkit",
}


def __getattr__(name: str) -> Any:
    """Look up attributes dynamically."""
    if name in _module_lookup:
        warnings.warn(
            (
                f"Import path `from langchain_ibm import {name}` is deprecated "
                "and may be removed in future. "
                f"Use `from langchain_ibm.agent_toolkits.utility import {name}` "
                "instead."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )
        module = importlib.import_module(_module_lookup[name])
        return getattr(module, name)
    error_msg = f"module {__name__} has no attribute {name}"
    raise AttributeError(error_msg)



"""Test IBM watsonx.ai Toolkit API wrapper."""

import os

import pytest
from ibm_watsonx_ai.wml_client_error import (  # type: ignore[import-untyped]
    WMLClientError,
)

from langchain_ibm.agent_toolkits.utility import WatsonxToolkit

os.environ.pop("WATSONX_APIKEY", None)
os.environ.pop("WATSONX_PROJECT_ID", None)


def test_initialize_watsonx_toolkit_without_url() -> None:
    pattern = r"(?=.*url)(?=.*WATSONX_URL)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxToolkit()


def test_initialize_watsonx_toolkit_cloud_only_url() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxToolkit(url="https://us-south.ml.cloud.ibm.com")


def test_initialize_watsonx_toolkit_with_deprecated_apikey() -> None:
    with (
        pytest.warns(
            DeprecationWarning,
            match="'apikey' parameter is deprecated; use 'api_key' instead.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxToolkit(
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_toolkit_with_api_key_and_apikey() -> None:
    with (
        pytest.warns(
            UserWarning,
            match="Both 'api_key' and deprecated 'apikey' were provided; "
            "'api_key' takes precedence.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxToolkit(
            url="https://us-south.ml.cloud.ibm.com",
            apikey="fake_apikey",
            api_key="fake_api_key",
        )


def test_initialize_watsonx_toolkit_cpd_without_all() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxToolkit(
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
        )


def test_initialize_watsonx_toolkit_cpd_only_password() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxToolkit(
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            password="test_password",  # noqa: S106
        )


def test_initialize_watsonx_toolkit_cpd_only_apikey() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxToolkit(
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_embeddings_cpd_deprecation_warning_with_instance_id() -> (
    None
):
    with (
        pytest.warns(
            DeprecationWarning, match="The `instance_id` parameter is deprecated"
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxToolkit(
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
            username="test_user",
            instance_id="openshift",
        )



import os

import pytest
from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]

from langchain_ibm import WatsonxEmbeddings

URL = "https://us-south.ml.cloud.ibm.com"
WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

MODEL = "ibm/granite-embedding-107m-multilingual"


@pytest.mark.skip("Until not released on us-south, should be skipped.")
class TestEmbeddingsGateway:
    documents = ("What is a generative ai?", "What is a loan and how does it works?")

    def test_embedding_model_gateway_init_credentials(self) -> None:
        watsonx_embedding = WatsonxEmbeddings(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        assert isinstance(watsonx_embedding, WatsonxEmbeddings)

        response = watsonx_embedding.embed_query(text=self.documents[0])
        assert response

    def test_embedding_model_gateway_init_client(self) -> None:
        credentials = Credentials(url=URL, api_key=WX_APIKEY)
        api_client = APIClient(credentials=credentials, project_id=WX_PROJECT_ID)

        watsonx_embedding = WatsonxEmbeddings(model=MODEL, watsonx_client=api_client)
        assert isinstance(watsonx_embedding, WatsonxEmbeddings)

        response = watsonx_embedding.embed_query(text=self.documents[0])
        assert response

    def test_embedding_model_gateway_embed_query(self) -> None:
        watsonx_embedding = WatsonxEmbeddings(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        generate_embedding = watsonx_embedding.embed_query(text=self.documents[0])
        assert isinstance(generate_embedding, list)
        assert isinstance(generate_embedding[0], float)

    async def test_embedding_model_gateway_aembed_query(self) -> None:
        watsonx_embedding = WatsonxEmbeddings(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        generate_embedding = await watsonx_embedding.aembed_query(
            text=self.documents[0]
        )
        assert isinstance(generate_embedding, list)
        assert isinstance(generate_embedding[0], float)

    def test_embedding_model_gateway_embed_documents(self) -> None:
        watsonx_embedding = WatsonxEmbeddings(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        generate_embedding = watsonx_embedding.embed_documents(
            texts=list(self.documents)
        )
        assert generate_embedding
        assert len(generate_embedding) == len(self.documents)
        assert all(isinstance(embedding, list) for embedding in generate_embedding)
        assert all(isinstance(embedding[0], float) for embedding in generate_embedding)
        assert len(generate_embedding[0]) > 0
        assert all(
            len(embedding) == len(generate_embedding[0])
            for embedding in generate_embedding
        )

    async def test_embedding_model_gateway_aembed_documents(self) -> None:
        watsonx_embedding = WatsonxEmbeddings(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        generate_embedding = await watsonx_embedding.aembed_documents(
            texts=list(self.documents)
        )
        assert generate_embedding
        assert len(generate_embedding) == len(self.documents)
        assert all(isinstance(embedding, list) for embedding in generate_embedding)
        assert all(isinstance(embedding[0], float) for embedding in generate_embedding)
        assert len(generate_embedding[0]) > 0
        assert all(
            len(embedding) == len(generate_embedding[0])
            for embedding in generate_embedding
        )



from typing import Any

from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from ibm_watsonx_ai.service_instance import (  # type: ignore[import-untyped]
    ServiceInstance,
)
from langchain_tests.unit_tests.embeddings import EmbeddingsUnitTests

from langchain_ibm import WatsonxEmbeddings

client = APIClient.__new__(APIClient)
client.CLOUD_PLATFORM_SPACES = True
client.ICP_PLATFORM_SPACES = True
credentials = Credentials(api_key="api_key")
client.credentials = credentials
client.service_instance = ServiceInstance.__new__(ServiceInstance)
client._httpx_client = None
client._async_httpx_client = None


class TestWatsonxEmbeddingsStandard(EmbeddingsUnitTests):
    @property
    def embeddings_class(self) -> type[WatsonxEmbeddings]:
        return WatsonxEmbeddings

    @property
    def embedding_model_params(self) -> dict[str, Any]:
        return {
            "model_id": "ibm/granite-13b-instruct-v2",
            "watsonx_client": client,
        }



"""Tools for interacting with a watsonx SQL databases via pyarrow.flight.FlightClient.

Based on the langchain_community.tools.sql_database.tool module.
"""

from typing import Any, cast

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import BaseMessage
from langchain_core.prompts import PromptTemplate
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import override

from langchain_ibm.utilities.sql_database import WatsonxSQLDatabase

QUERY_CHECKER = """
{query}
Double check the query above for common mistakes, including:
- Using NOT IN with NULL values
- Using UNION when UNION ALL should have been used
- Using BETWEEN for exclusive ranges
- Data type mismatch in predicates
- Properly quoting identifiers
- Using the correct number of arguments for functions
- Casting to the correct data type
- Using the proper columns for joins
- Make sure that schema name `{schema}` is added to the table name, e.g. {schema}.table1

If there are any of the above mistakes, rewrite the query. If there are no mistakes, just reproduce the original query.

Output the final SQL query only.

SQL Query: """  # noqa: E501


class BaseSQLDatabaseTool(BaseModel):
    """Base tool for interacting with a SQL database."""

    db: WatsonxSQLDatabase = Field(exclude=True)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class _QuerySQLDatabaseToolInput(BaseModel):
    query: str = Field(..., description="A detailed and correct SQL query.")


class QuerySQLDatabaseTool(BaseSQLDatabaseTool, BaseTool):
    """Tool for querying a SQL database."""

    name: str = "sql_db_query"
    description: str = """
    Execute a SQL query against the database and get back the result.
    If the query is not correct, an error message will be returned.
    If an error is returned, rewrite the query, check the query correctness,
    and try again.
    """
    args_schema: type[BaseModel] = _QuerySQLDatabaseToolInput

    @override
    def _run(
        self,
        query: str,
        run_manager: CallbackManagerForToolRun | None = None,
    ) -> str:
        """Execute the query, return the results or an error message."""
        return self.db.run_no_throw(query)


class _InfoSQLDatabaseToolInput(BaseModel):
    table_names: str = Field(
        ...,
        description=(
            "A comma-separated list of the table names "
            "for which to return the schema. "
            "Example input: 'table1, table2, table3'"
        ),
    )


class InfoSQLDatabaseTool(BaseSQLDatabaseTool, BaseTool):
    """Tool for getting metadata about a SQL database."""

    name: str = "sql_db_schema"
    description: str = "Get the schema and sample rows for the specified SQL tables."
    args_schema: type[BaseModel] = _InfoSQLDatabaseToolInput

    @override
    def _run(
        self,
        table_names: str,
        run_manager: CallbackManagerForToolRun | None = None,
    ) -> str:
        """Get the schema for tables in a comma-separated list."""
        return self.db.get_table_info_no_throw(
            [t.strip() for t in table_names.split(",")],
        )


class _ListSQLDatabaseToolInput(BaseModel):
    tool_input: str = Field("", description="An empty string")


class ListSQLDatabaseTool(BaseSQLDatabaseTool, BaseTool):
    """Tool for getting tables names."""

    name: str = "sql_db_list_tables"
    description: str = (
        "Input is an empty string, output is a comma-separated list "
        "of tables in the database."
    )
    args_schema: type[BaseModel] = _ListSQLDatabaseToolInput

    @override
    def _run(
        self,
        tool_input: str = "",
        run_manager: CallbackManagerForToolRun | None = None,
    ) -> str:
        """Get a comma-separated list of table names."""
        return ", ".join(self.db.get_usable_table_names())


class _QuerySQLCheckerToolInput(BaseModel):
    query: str = Field(..., description="A detailed and SQL query to be checked.")


class QuerySQLCheckerTool(BaseSQLDatabaseTool, BaseTool):
    """Use an LLM to check if a query is correct."""

    template: str = QUERY_CHECKER
    llm: BaseLanguageModel[Any]
    llm_chain: Any = Field(init=False)
    name: str = "sql_db_query_checker"
    description: str = """
    Use this tool to double check if your query is correct before executing it.
    Always use this tool before executing a query with sql_db_query!
    """
    args_schema: type[BaseModel] = _QuerySQLCheckerToolInput

    @model_validator(mode="before")
    @classmethod
    def initialize_llm_chain(cls, values: dict[str, Any]) -> Any:
        """Initialize llm chain."""
        if "llm_chain" not in values:
            prompt = PromptTemplate(
                template=QUERY_CHECKER,
                input_variables=["query", "schema"],
            )
            llm = cast("BaseLanguageModel[Any]", values.get("llm"))

            values["llm_chain"] = prompt | llm

        if values["llm_chain"].first.input_variables != ["query", "schema"]:
            error_msg = (
                "LLM chain for QueryCheckerTool must have input variables "
                "['query', 'schema']",
            )
            raise ValueError(error_msg)

        return values

    def _run(
        self,
        query: str,
        run_manager: CallbackManagerForToolRun | None = None,
    ) -> str:
        """Use the LLM to check the query."""
        resp = self.llm_chain.invoke(
            {"query": query, "schema": self.db.schema},
            callbacks=run_manager.get_child() if run_manager else None,
        )
        if isinstance(resp, BaseMessage):
            return str(resp.content)
        return cast("str", resp)

    async def _arun(
        self,
        query: str,
        run_manager: AsyncCallbackManagerForToolRun | None = None,
    ) -> str:
        resp = await self.llm_chain.ainvoke(
            {"query": query, "schema": self.db.schema},
            callbacks=run_manager.get_child() if run_manager else None,
        )
        if isinstance(resp, BaseMessage):
            return str(resp.content)
        return cast("str", resp)



from unittest.mock import Mock

import pytest
from pyarrow import flight  # type: ignore[import-untyped]

from langchain_ibm.agent_toolkits.sql.tool import (
    InfoSQLDatabaseTool,
    ListSQLDatabaseTool,
    QuerySQLDatabaseTool,
)
from langchain_ibm.utilities.sql_database import WatsonxSQLDatabase


@pytest.fixture
def mock_db() -> Mock:
    return Mock(spec=WatsonxSQLDatabase)


### QuerySQLDatabaseTool


def test_query_tool_run_with_valid_query(mock_db: Mock) -> None:
    """Test running a valid query."""
    mock_db.run_no_throw.return_value = "Valid query result"
    tool = QuerySQLDatabaseTool(db=mock_db)

    query = "SELECT * FROM table"
    result = tool._run(query, None)
    assert result == "Valid query result"
    mock_db.run_no_throw.assert_called_once_with(query)


def test_query_tool_run_with_invalid_query(mock_db: Mock) -> None:
    """Test running an invalid query."""

    def mock_run_no_throw() -> None:
        error_msg = "Invalid query"
        raise flight.FlightError(error_msg)

    mock_db.run.side_effect = mock_run_no_throw
    tool = QuerySQLDatabaseTool(db=mock_db)

    query = "SELECT * FROM non_existent_table"
    result = tool._run(query, None)
    assert result.startswith("Error"), "Expected an error message"
    mock_db.run_no_throw.assert_called_once_with(query)


### InfoSQLDatabaseTool


def test_info_tool_run_with_valid_query(mock_db: Mock) -> None:
    """Test running a valid query."""
    mock_db.get_table_info_no_throw.return_value = "schema_info"
    tool = InfoSQLDatabaseTool(db=mock_db)

    result = tool._run("table1,table2", None)

    mock_db.get_table_info_no_throw.assert_called_once_with(["table1", "table2"])
    assert result == "schema_info"


def test_info_tool_run_with_invalid_query(mock_db: Mock) -> None:
    """Test running an invalid query."""

    def mock_get_table_info() -> None:
        error_msg = "Table not Found"
        raise flight.FlightError(error_msg)

    mock_db.get_table_info.side_effect = mock_get_table_info
    tool = InfoSQLDatabaseTool(db=mock_db)

    result = tool._run("tableX", None)

    mock_db.get_table_info_no_throw.assert_called_once_with(["tableX"])
    assert result.startswith("Error"), "Expected an error message"


### ListSQLDatabaseTool


def test_list_tool_run_with_valid_query(mock_db: Mock) -> None:
    """Test running a valid query."""
    mock_db.get_usable_table_names.return_value = ["table1", "table2"]
    tool = ListSQLDatabaseTool(db=mock_db)

    result = tool._run()

    mock_db.get_usable_table_names.assert_called_once_with()
    assert result == "table1, table2"



"""Test WatsonxEmbeddings.

You'll need to set WATSONX_APIKEY and WATSONX_PROJECT_ID environment variables.
"""

import os
from typing import Any

import pytest
from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.embeddings import (  # type: ignore[import-untyped]
    Embeddings,
)
from ibm_watsonx_ai.metanames import (  # type: ignore[import-untyped]
    EmbedTextParamsMetaNames,
)

from langchain_ibm import WatsonxEmbeddings

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"
MODEL_ID = "ibm/granite-embedding-278m-multilingual"

DOCUMENTS = ["What is a generative ai?", "What is a loan and how does it works?"]

CREATE_WATSONX_EMBEDDINGS_INIT_PARAMETERS = [
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only api_key",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only apikey",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="api_key and apikey",
    ),
]


@pytest.mark.parametrize("init_data", CREATE_WATSONX_EMBEDDINGS_INIT_PARAMETERS)
def test_watsonx_embeddings_init(init_data: dict[str, Any]) -> None:
    watsonx_embedding = WatsonxEmbeddings(**init_data)

    generate_embedding = watsonx_embedding.embed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_init_with_client() -> None:
    watsonx_client = APIClient(
        credentials={
            "url": "https://us-south.ml.cloud.ibm.com",
            "apikey": WX_APIKEY,
        },
        project_id=WX_PROJECT_ID,
    )
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID, watsonx_client=watsonx_client
    )
    generate_embedding = watsonx_embedding.embed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_init_with_embeddings() -> None:
    watsonx_client = APIClient(
        credentials={
            "url": "https://us-south.ml.cloud.ibm.com",
            "apikey": WX_APIKEY,
        },
        project_id=WX_PROJECT_ID,
    )
    embedding = Embeddings(api_client=watsonx_client, model_id=MODEL_ID)

    watsonx_embedding = WatsonxEmbeddings(
        watsonx_embed=embedding,
    )
    generate_embedding = watsonx_embedding.embed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_01_generate_embed_documents() -> None:
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = watsonx_embedding.embed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


async def test_01_generate_aembed_documents() -> None:
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = await watsonx_embedding.aembed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_02_generate_embed_documents_with_param() -> None:
    embed_params = {
        EmbedTextParamsMetaNames.TRUNCATE_INPUT_TOKENS: 3,
    }
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=embed_params,
    )
    generate_embedding = watsonx_embedding.embed_documents(texts=DOCUMENTS)
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_03_generate_embed_documents_with_param_in_method() -> None:
    embed_params = {
        EmbedTextParamsMetaNames.TRUNCATE_INPUT_TOKENS: 3,
    }
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = watsonx_embedding.embed_documents(
        texts=DOCUMENTS, params=embed_params
    )
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_04_generate_embed_documents_with_param_and_concurrency_limit() -> None:
    embed_params = {
        EmbedTextParamsMetaNames.TRUNCATE_INPUT_TOKENS: 3,
    }
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=embed_params,
    )
    generate_embedding = watsonx_embedding.embed_documents(
        texts=DOCUMENTS, concurrency_limit=9
    )
    assert len(generate_embedding) == len(DOCUMENTS)
    assert all(isinstance(el, float) for el in generate_embedding[0])


def test_10_generate_embed_query() -> None:
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = watsonx_embedding.embed_query(text=DOCUMENTS[0])
    assert isinstance(generate_embedding, list), "embed_query should return a list"

    assert len(generate_embedding) > 0, "embed_query returned an empty list"

    assert isinstance(generate_embedding[0], float), (
        "First element of embed_query should be a float"
    )


async def test_10_generate_aembed_query() -> None:
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = await watsonx_embedding.aembed_query(text=DOCUMENTS[0])
    assert isinstance(generate_embedding, list), "embed_query should return a list"

    assert len(generate_embedding) > 0, "embed_query returned an empty list"

    assert isinstance(generate_embedding[0], float), (
        "First element of embed_query should be a float"
    )


def test_11_generate_embed_query_with_params() -> None:
    embed_params = {
        EmbedTextParamsMetaNames.TRUNCATE_INPUT_TOKENS: 3,
    }
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    generate_embedding = watsonx_embedding.embed_query(
        text=DOCUMENTS[0], params=embed_params
    )
    assert isinstance(generate_embedding, list), "embed_query should return a list"

    assert len(generate_embedding) > 0, "embed_query returned an empty list"

    assert isinstance(generate_embedding[0], float), (
        "First element of embed_query should be a float"
    )


def test_12_generate_embed_query_with_params_and_concurrency_limit() -> None:
    embed_params = {
        EmbedTextParamsMetaNames.TRUNCATE_INPUT_TOKENS: 3,
    }
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=embed_params,
    )
    generate_embedding = watsonx_embedding.embed_query(
        text=DOCUMENTS[0], concurrency_limit=9
    )
    assert isinstance(generate_embedding, list), "embed_query should return a list"

    assert len(generate_embedding) > 0, "embed_query returned an empty list"

    assert isinstance(generate_embedding[0], float), (
        "First element of embed_query should be a float"
    )


def test_20_generate_embed_query_with_client_initialization() -> None:
    watsonx_client = APIClient(
        credentials={
            "url": URL,
            "apikey": WX_APIKEY,
        }
    )
    watsonx_embedding = WatsonxEmbeddings(
        model_id=MODEL_ID, project_id=WX_PROJECT_ID, watsonx_client=watsonx_client
    )
    generate_embedding = watsonx_embedding.embed_query(text=DOCUMENTS[0])

    assert isinstance(generate_embedding, list), "embed_query should return a list"

    assert len(generate_embedding) > 0, "embed_query returned an empty list"
    assert isinstance(generate_embedding[0], float), "First element should be a float"
